# Copyright 2021-2022, Barcelona Supercomputing Center (BSC), Spain
#
# This software was partially supported by the EuroHPC-funded project ADMIRE
#   (Project ID: 956748, https://www.admire-eurohpc.eu).
#
# This file is part of rpcc.
#
# rpcc is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
#
# rpcc is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License along with rpcc.  If not, see
# <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import itertools
import pathlib
import re
from typing import Optional, Union, Callable, Iterable, Tuple, List

from loguru import logger

from rpcc.ir import Argument, ReturnValue, RemoteProcedure, IRTree, ConditionalDefinition
from rpcc.transformers.cxx17.types import (
    CxxType, Integer, Double, String, Object, NameSpace, ReferenceType, Identifier, ConstLValueReference,
    RValueReference
)


class CxxParameterListFormatter:
    """Format a list of C++ function parameters."""

    sep: str = ", "

    def format(self, params: Callable[..., Iterable[Identifier]]) -> str:
        """Generate C++ code for a parameter list.

        :param params: A function that generates the parameters that should appear in the list.
        :return: A string with the generated C++ code.

        Example:
            >>> CxxParameterListFormatter().format(
            >>>     lambda: iter([
            >>>         Identifier("a", Integer(32, False)),
            >>>         Identifier("b", Double()),
            >>>         ConstLValueReference("c", String())
            >>> ]))
            'int a, double b, const std::string& c'
        """
        return self.sep.join(p.as_declarator() for p in params())


class CxxInitializerListFormatter:
    """Format a C++ member initializer list."""

    sep: str = ", "

    @staticmethod
    def _format_init_expr(member: Identifier, param: Identifier) -> str:
        if not member:
            return ""

        if not param:
            return f"{member}()"

        return f"{member}({param.as_member_of()})"

    def format(self, params: Callable[..., Iterable[Identifier]], members: Callable[..., Iterable[Identifier]]) -> str:

        init_list = self.sep.join(
            itertools.starmap(CxxInitializerListFormatter._format_init_expr,
                              itertools.zip_longest(members(), params())))

        if init_list:
            return f": {init_list}"
        return init_list


class CxxConstructorFormatter:
    """Format a C++ constructor function."""

    classname: str
    template: str = """\
{decl_specifier}{classname}({params}){initializers} {body}\n"""

    def __init__(self, classname: str) -> None:
        self.classname = classname

    def format(self, explicit: bool, ctor_params: Callable[..., Iterable[Identifier]],
               class_members: Callable[..., Iterable[Identifier]], body=None) -> str:
        return self.template.format(
            decl_specifier="explicit\n" if explicit else "",
            classname=self.classname,
            params=CxxParameterListFormatter().format(ctor_params),
            initializers=CxxInitializerListFormatter().format(ctor_params, class_members),
            body=body or "{ }"
        )


class CxxConvertingConstructorFormatter:
    """Format a C++ converting constructor function."""

    classname: str
    template: str = """\
{decl_specifier}{classname}({params}){initializers} {body}\n"""

    def __init__(self, classname: str) -> None:
        """ Create a `ConvertingConstructorFormatter` given a class name.

        :param classname: The output class name created by the constructor.
        """
        self.classname = classname

    def format(self, explicit: bool, class_members: Callable[..., Iterable[Identifier]], ctor_param: Identifier,
               member_values: Callable[..., Iterable[Identifier]], body: object = None) -> str:
        """Generate a C++ converting constructor.

        :param explicit: Whether the constructor should be explicit or not.
        :param ctor_param: The constructor parameter that should be converted to `classname`.
        :param class_members: The class members of `classname` that should be initialized when converting.
        :param member_values: The values from `ctor_param` that should be used to initialize `class_members`
        :param body: The body of the constructor.
        :return: A string with the generated code.
        """
        init_list = CxxInitializerListFormatter().format(class_members, member_values)

        if not init_list:
            body = f"{{ (void) {ctor_param}; }}"

        return self.template.format(
            decl_specifier="explicit\n" if explicit else "",
            classname=self.classname,
            params=CxxParameterListFormatter().format(lambda: iter([ctor_param])),
            initializers=init_list,
            body=body or "{ }"
        )


class CxxConvertingOperatorFormatter:
    """Format a C++ converting operator."""

    template: str = """\
{decl_specifier}operator {conversion_type}() {{
    return {{ {converting_operator_retval} }};
}}\n"""

    def format(self, explicit: bool, conversion_type: CxxType, conversion_func: Callable[[Identifier], str],
               values: Callable[..., Iterable[Identifier]]) -> str:
        return self.template.format(
            decl_specifier="explicit\n" if explicit else "",
            conversion_type=conversion_type,
            converting_operator_retval=CxxBracedInitializerListFormatter().format(conversion_func, values),
        )


class CxxFunctionFormatter:
    """Format a C++ function."""

    template: str = """\
{return_type} {func_name}({parameter_list}) {cv} {body}\n"""

    def format(self, return_type: CxxType, func_name: str, func_params: Callable[..., Iterable[Identifier]],
               is_const: bool,
               body: Optional[str] = None):
        return self.template.format(
            return_type=return_type,
            func_name=func_name,
            parameter_list=CxxParameterListFormatter().format(func_params),
            cv="const" if is_const else "",
            body=body or "{ }"
        )


class CxxBracedInitializerListFormatter:
    """Format a C++ brace-enclosed initializer list."""

    mapping_func: Callable
    sep: str = ",\n"

    def format(self, func: Callable[[Identifier], str], values: Callable[..., Iterable[Identifier]]):
        return self.sep.join(map(func, values()))


HERMES_NAMESPACE = NameSpace("hermes::detail")
HG_VOID = Object("hg_void_t", HERMES_NAMESPACE)

INPUT_CLASS_NAME = "input"
OUTPUT_CLASS_NAME = "output"
CONSTRUCTOR_VISIBILITY = "public"
OPERATOR_VISIBILITY = "public"
GETTER_VISIBILITY = "public"
MEMBER_VISIBILITY = "private"

RpcVar = Union[Argument, ReturnValue]


def as_members(args: Iterable[RpcVar], owner: Optional[Identifier] = None) -> Iterable[Identifier]:
    """Given a list of RPC variables, generate a sequence of appropriately formatted C++ class members.

    :param args: The list of RPC variables
    :param owner: An optional class instance owner for the member.
    """
    for a in args:
        yield Identifier("m_" + a.id, a.typeinfo, owner)


def as_function_parameters(args: Iterable[RpcVar]) -> Iterable[Identifier]:
    """Given a list of RPC variables, generate a sequence of appropriately formatted C++ function parameters.
    Fundamental types are formatted as by-value types, while compound types are formatted as const l-value references.

    :param args: The list of RPC variables
    """
    for a in args:
        if a.typeinfo.is_fundamental():
            yield Identifier(a.id, a.typeinfo)
        else:
            yield ConstLValueReference(a.id, a.typeinfo)


class _InnerClassFormatter:
    rpcname: str
    classname: str
    hg_type: Object
    params: Iterable[RpcVar]
    template: str = """\
class {classname} {{

    template <typename ExecutionContext>
    friend hg_return_t hermes::detail::post_to_mercury(ExecutionContext*);

    {constructors}

    {operators}

    {getters}

    {members}
}};\n"""

    def __init__(self, rpcname: str, classname: str, hg_type: Object):
        self.rpcname = rpcname
        self.classname = classname
        self.hg_type = hg_type

    def format(self, rpc_vars: Iterable[RpcVar]):
        return self.template.format(
            rpc_name=self.rpcname,
            classname=self.classname,
            constructors=self._format_constructors(CONSTRUCTOR_VISIBILITY, rpc_vars, sep="\n"),
            operators=self._format_operators(OPERATOR_VISIBILITY, rpc_vars, sep="\n"),
            getters=self._format_getters(GETTER_VISIBILITY, rpc_vars, sep="\n"),
            members=self._format_members(MEMBER_VISIBILITY, rpc_vars, sep="\n")
        )

    def _format_constructors(self, visibility: str, args: Iterable[RpcVar], sep: str) -> str:
        """Generate C++ code for the set of constructors required by this RPC inner class.

        The following constructors are generated:
            - A constructor for default-initialized members
            - A parameterized constructor (unless `args` contains no variables)
            - A copy constructor
            - A move constructor
            - A converting constructor

        :param visibility: The visibility that should be used for the generated constructors. Valid values are `public`,
        `private`, or `protected`.
        :param args: The RPC variables (arguments or return variables) that define this RPC.
        :param sep: The separator that should be used to separate constructor code.
        :return: A string with the generated C++ code.
        """

        if visibility not in ["public", "private", "protected"]:
            raise ValueError(f"Unexpected value '{visibility}' for constructor visibility")

        # constructor for default-initialized members
        logger.info(f"formatting constructor for '{self.classname}' default-initialized members")
        constructors = [
            CxxConstructorFormatter(self.classname).format(
                explicit=False,
                ctor_params=lambda: iter([]),
                class_members=lambda: as_members(args)
            )]
        logger.info(f"output:\n'{constructors[-1]}'")

        # parameterized constructor (only if args is not empty)
        if args:
            logger.info(f"formatting parameterized constructor for '{self.classname}'")
            constructors.append(
                CxxConstructorFormatter(self.classname).format(
                    explicit=False,
                    ctor_params=lambda: as_function_parameters(args),
                    class_members=lambda: as_members(args)
                ))
            logger.info(f"output:\n'{constructors[-1]}'")

        # move constructor
        logger.info(f"formatting move constructor for '{self.classname}'")
        constructors.append(
            CxxConstructorFormatter(self.classname).format(
                explicit=False,
                ctor_params=lambda: iter([RValueReference("rhs", Object(self.classname))]),
                class_members=lambda: iter([]),
                body="= default;"
            )
        )
        logger.info(f"output:\n'{constructors[-1]}'")

        # copy constructor
        logger.info(f"formatting copy constructor for '{self.classname}'")
        constructors.append(
            CxxConstructorFormatter(self.classname).format(
                explicit=False,
                ctor_params=lambda: iter([ConstLValueReference("other", Object(self.classname))]),
                class_members=lambda: iter([]),
                body="= default;"
            )
        )
        logger.info(f"output:\n'{constructors[-1]}'")

        # converting constructor
        logger.info(f"formatting converting constructor for '{self.classname}'")
        owner = ConstLValueReference("other", self.hg_type)
        constructors.append(
            CxxConvertingConstructorFormatter(self.classname).format(
                explicit=True,
                class_members=lambda: as_members(args, owner),
                ctor_param=owner,
                member_values=lambda: as_members(args))
        )
        logger.info(f"output:\n'{constructors[-1]}'")

        return f"{visibility}:\n" + sep.join(constructors)

    def _format_operators(self, visibility: str, args: Iterable[RpcVar], sep: str) -> str:
        """Generate C++ code for the set of operators required by this RPC inner class.

        The following operators are generated:
            - A move-assignment operator
            - An assignment operator
            - A conversion operator

        :param visibility: The visibility that should be used for the generated constructors. Valid values are `public`,
        `private`, or `protected`.
        :param args: The RPC variables (arguments or return variables) that define this RPC.
        :param sep: The separator that should be used to separate constructor code.
        :return: A string with the generated C++ code.
        """

        # move-assignment operator
        logger.info(f"formatting move-assignment operator for '{self.classname}'")
        operators = [
            CxxFunctionFormatter().format(
                return_type=ReferenceType(Object(self.classname)),
                func_name="operator=",
                func_params=lambda: iter([RValueReference("rhs", Object(self.classname))]),
                is_const=False,
                body="= default;"
            )
        ]
        logger.info(f"output:\n'{operators[-1]}'")

        # assignment operator
        logger.info(f"formatting assignment operator for '{self.classname}'")
        operators.append(
            CxxFunctionFormatter().format(
                return_type=ReferenceType(Object(self.classname)),
                func_name="operator=",
                func_params=lambda: iter([ConstLValueReference("other", Object(self.classname))]),
                is_const=False,
                body="= default;"
            )
        )
        logger.info(f"output:\n'{operators[-1]}'")

        # conversion operator
        logger.info(f"formatting converting operator for '{self.classname}'")
        operators.append(
            CxxConvertingOperatorFormatter().format(
                explicit=True,
                conversion_type=self.hg_type,
                conversion_func=lambda ident: ident.as_hg_conversion(),
                values=lambda: as_members(args),
            )
        )
        logger.info(f"output:\n'{operators[-1]}'")

        return f"{visibility}:\n" + sep.join(operators)

    def _format_getters(self, visibility: str, args: Iterable[RpcVar], sep) -> str:
        return ((f"{visibility}:\n" if args else "") +
                sep.join(map(lambda m:
                             CxxFunctionFormatter().format(
                                 return_type=m.typeinfo,
                                 func_name=m.id.replace('m_', ''),
                                 func_params=lambda: iter([]),
                                 is_const=True,
                                 body=f"{{ return {m.id}; }}"
                             ), as_members(args))
                         )
                )

    def _format_members(self, visibility: str, args: Iterable[RpcVar], sep) -> str:
        return ((f"{visibility}:\n" if args else "") +
                sep.join(f"{m.as_declarator()};" for m in as_members(args))
                )


class _RemoteProcedureFormatter:
    rpc: RemoteProcedure
    hg_input_type: Object = HG_VOID
    hg_output_type: Object = HG_VOID
    decl_template: str = """\
//==============================================================================
// definitions for {rpc_name}
namespace hermes::detail {{

// Generate Mercury types and serialization functions (field names match
// those defined in {rpc_name}::input and {rpc_name}::output). These
// definitions are internal and should not be used directly. Classes
// {rpc_name}::input and {rpc_name}::output are provided for public use.
{hg_input_type_definition}

{hg_output_type_definition}

}} // namespace hermes::detail

struct {rpc_name} {{

// forward declarations of public input/output types for this RPC
class input;

class output;

// traits used so that the engine knows what to do with the RPC
using self_type = {rpc_name};
using handle_type = hermes::rpc_handle<self_type>;
using input_type = input;
using output_type = output;
using mercury_input_type = {hg_input_type};
using mercury_output_type = {hg_output_type};

// RPC public identifier
constexpr static const uint64_t public_id = {rpc_id};

// RPC internal Mercury identifier
constexpr static const hg_id_t mercury_id = public_id;

// RPC name
constexpr static const auto name = "{rpc_name}";

// requires response?
constexpr static const auto requires_response = true;

// Mercury callback to serialize input arguments
constexpr static const auto mercury_in_proc_cb = 
    {hg_input_type_callback};

// Mercury callback to serialize output arguments
constexpr static const auto mercury_out_proc_cb = 
    {hg_output_type_callback};

{rpc_args}

{rpc_retval}

}};\n"""

    def_template: str = """\
(void) registered_requests().add <{rpc_name}>();"""

    def __init__(self, rpc: RemoteProcedure):
        self.rpc = rpc
        if len(self.rpc.args):
            self.hg_input_type = Object(f"{self.rpc.name}_in_t", HERMES_NAMESPACE)
        if len(self.rpc.retvals):
            self.hg_output_type = Object(f"{self.rpc.name}_out_t", HERMES_NAMESPACE)

    def _format_hg_type_definition(self, name: str, fields: Iterable[RpcVar]) -> str:

        # hg_void_t is generated elsewhere
        if name == "hg_void_t":
            return ""

        return (f"MERCURY_GEN_PROC(\n" +
                f"{name},\n" +
                "".join(f"(({f.typeinfo.hg_type()}) ({f.id}))" for f in fields) +
                ")"
                )

    def _format_hg_type_callback(self, classname: str, fields: List[RpcVar]) -> str:
        if not len(fields):
            return "hermes::detail::hg_proc_void_t"
        return f"HG_GEN_PROC_NAME({classname})"

    def format_declaration(self):
        logger.info(f"formatting rpc '{self.rpc.name}'")

        expr = self.rpc.include_if_expr

        pre = f"{expr.start_keyword} {expr.symbol}\n\n" if expr else ""
        post = f"{expr.end_keyword} // {expr.symbol}\n\n" if expr else ""

        return (pre +
                self.decl_template.format(
                    rpc_id=self.rpc.id,
                    rpc_name=self.rpc.name,
                    hg_input_type=self.hg_input_type,
                    hg_input_type_definition=self._format_hg_type_definition(self.hg_input_type.classname,
                                                                             self.rpc.args),
                    hg_input_type_callback=self._format_hg_type_callback(self.hg_input_type.classname, self.rpc.args),
                    hg_output_type=self.hg_output_type,
                    hg_output_type_definition=self._format_hg_type_definition(self.hg_output_type.classname,
                                                                              self.rpc.retvals),
                    hg_output_type_callback=self._format_hg_type_callback(self.hg_output_type.classname,
                                                                          self.rpc.retvals),
                    rpc_args=_InnerClassFormatter(
                        self.rpc.name,
                        INPUT_CLASS_NAME,
                        self.hg_input_type
                    ).format(self.rpc.args),
                    rpc_retval=_InnerClassFormatter(
                        self.rpc.name,
                        OUTPUT_CLASS_NAME,
                        self.hg_output_type
                    ).format(self.rpc.retvals),
                ) + "\n" +
                post)

    def format_definition(self, namespace: Optional[str]) -> str:
        rpc_fqname = self.rpc.name
        if namespace:
            rpc_fqname = '::'.join([namespace, rpc_fqname])
        return self.def_template.format(rpc_name=rpc_fqname)


class _HppHeaderFormatter:
    template: str = """\
{copyright_text}
// Generated by rpcc. DO NOT EDIT!
// source: {input_file}

#ifndef {header_guard_name}
#define {header_guard_name}

// C includes
#include <mercury.h>
#include <mercury_proc_string.h>
#include <mercury_macros.h>

// C++ includes
#include <string>

// hermes includes
#include <hermes.hpp>

#ifndef HG_GEN_PROC_NAME
#define HG_GEN_PROC_NAME(struct_type_name) \
    hermes::detail::hg_proc_ ## struct_type_name
#endif

// forward declarations
namespace hermes::detail {{

template <typename ExecutionContext>
hg_return_t post_to_mercury(ExecutionContext* ctx);

struct hg_void_t {{}};

static HG_INLINE hg_return_t
hg_proc_void_t(hg_proc_t proc, void* data) {{
    (void) proc;
    (void) data;

    return HG_SUCCESS;
}}

}} // namespace hermes::detail
"""

    def format(self, input_file: str, copyright_text: str, header_guard: str) -> str:
        return self.template.format(
            header_guard_name=header_guard,
            copyright_text=copyright_text,
            input_file=input_file)


class _HppFooterFormatter:
    template: str = """\
#undef HG_GEN_PROC_NAME

#endif // {header_guard_name}
"""

    def format(self, header_guard: str) -> str:
        return self.template.format(
            header_guard_name=header_guard,
        )


def _format_cxx_namespace(name: str, entry: bool) -> str:
    if entry:
        return f"namespace {name} {{\n"
    return f"}} // namespace {name}\n"


def _format_cxx_conditional_expression(expr: ConditionalDefinition, entry: bool) -> str:
    if entry:
        return f"{expr.start_keyword} {expr.symbol}"
    return f"{expr.end_keyword} // {expr.symbol}"


class _CxxHppFileFormatter:
    @staticmethod
    def _format_cxx_header_guard(input_file: pathlib.Path) -> str:
        prefix = input_file.name
        extensions = input_file.suffixes

        if extensions:
            prefix = re.compile('|'.join(map(re.escape, extensions))).sub('', prefix)

        return f"RPCC_{prefix.upper()}_INCLUDED_HPP"

    @staticmethod
    def format(input_file: pathlib.Path, copyright_text: str, namespace: Optional[str],
               rpcs: Iterable[RemoteProcedure]) -> str:
        return (
                _HppHeaderFormatter().format(
                    input_file=str(input_file),
                    copyright_text=copyright_text,
                    header_guard=_CxxHppFileFormatter._format_cxx_header_guard(input_file)
                ) + "\n" +
                (_format_cxx_namespace(namespace, entry=True) + "\n" if namespace else "") +
                "\n".join(_RemoteProcedureFormatter(rpc).format_declaration() for rpc in rpcs) + "\n" +
                (_format_cxx_namespace(namespace, entry=False) + "\n" if namespace else "") +
                _HppFooterFormatter().format(
                    header_guard=_CxxHppFileFormatter._format_cxx_header_guard(input_file)
                )
        )


class _CppHeaderFormatter:
    template: str = """\
{copyright_text}
// Generated by rpcc. DO NOT EDIT!
// source: {input_file}

// hermes includes
#include <hermes.hpp>
#include "{hpp_file}"

namespace hermes::detail {{

//==============================================================================
// register request types so that they can be used by users and the engine
//
"""

    def format(self, input_file: str, copyright_text: str, hpp_file: str) -> str:
        return self.template.format(
            copyright_text=copyright_text,
            input_file=input_file,
            hpp_file=hpp_file)


class _CppFooterFormatter:
    def format(self) -> str:
        return "\n} // namespace hermes::detail"


class _CxxCppFileFormatter:
    @staticmethod
    def format(input_file: pathlib.Path, copyright_text: str, hpp_file: str, namespace: Optional[str],
               rpcs: Iterable[RemoteProcedure]) -> str:
        return (
                _CppHeaderFormatter().format(
                    input_file=str(input_file),
                    copyright_text=copyright_text,
                    hpp_file=hpp_file
                ) +
                "\n".join(_RemoteProcedureFormatter(rpc).format_definition(namespace) for rpc in rpcs) + "\n" +
                _CppFooterFormatter().format()
        )


class FileFormatter:
    input_file: pathlib.Path
    copyright_file: pathlib.Path
    header_output_path: pathlib.Path
    impl_output_path: pathlib.Path

    def __init__(self, input_file: pathlib.Path, copyright_file: pathlib.Path, header_output_path: pathlib.Path,
                 impl_output_path: pathlib.Path):
        self.input_file = input_file
        self.copyright_file = copyright_file
        self.header_output_path = header_output_path
        self.impl_output_path = impl_output_path

    def format(self, tree: IRTree) -> Tuple[str, str]:
        # since C++17 allows collapsing namespaces, we can just replace '.' with '::' in the package name
        ns = tree.package.replace('.', '::') if tree.package else None

        copyright_text = self.copyright_file.read_text() if self.copyright_file else ""

        hpp_text = _CxxHppFileFormatter.format(
            input_file=self.input_file,
            copyright_text=copyright_text,
            namespace=ns,
            rpcs=tree.rpcs)

        cpp_text = _CxxCppFileFormatter.format(
            input_file=self.input_file,
            copyright_text=copyright_text,
            hpp_file=self.header_output_path.name,
            namespace=ns,
            rpcs=tree.rpcs)

        return hpp_text, cpp_text



# Copyright 2021-2022, Barcelona Supercomputing Center (BSC), Spain
#
# This software was partially supported by the EuroHPC-funded project ADMIRE
#   (Project ID: 956748, https://www.admire-eurohpc.eu).
#
# This file is part of rpcc.
#
# rpcc is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
#
# rpcc is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License along with rpcc.  If not, see
# <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import argparse
import logging
import sys
from pathlib import Path
from rpcc.version import __version__ as rpcc_version
from rpcc.console import console


class _SetVerbosity(argparse.Action):
    def __init__(self, option_strings, dest, const=None, default=None,
                 required=False, help=None, metavar=None):
        super().__init__(option_strings=option_strings,
                         dest=dest,
                         nargs=0,
                         const=const,
                         default=default,
                         required=required,
                         help=help)

        self.count = 0

    def __call__(self, parser, namespace, values, option_string=None):
        levels = [
            logging.CRITICAL,
            logging.ERROR,
            logging.WARNING,
            25,  # LOGURU_SUCCESS_NO
            logging.INFO,
            logging.DEBUG,
            5  # LOGURU_TRACE_NO
        ]

        setattr(namespace, self.dest, levels[self.count])
        self.count = min(self.count + 1, len(levels) - 1)


class _PrintVersion(argparse.Action):
    def __init__(self, option_strings, dest, const=None, default=None,
                 required=False, help=None, metavar=None):
        super().__init__(option_strings=option_strings,
                         dest=dest,
                         nargs=0,
                         const=const,
                         default=default,
                         required=required,
                         help=help)

    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, self.const)
        console.print(f"rpcc {rpcc_version}")
        sys.exit(0)


def parse_args(args) -> argparse.Namespace:
    parser = argparse.ArgumentParser(prog="rpcc",
                                     description="Parse RPC_PROTO_FILE and generate C++ output based on the options "
                                                 "given.")
    parser.add_argument(
        "rpc_proto_file",
        type=Path,
        metavar="RPC_PROTO_FILE",
        help="The file containing the specifications for the RPCs."
    )

    parser.add_argument(
        "--output-mode",
        choices=["hermes"],
        dest="output_mode",
        metavar="LIBRARY",
        default="hermes",
        help="Set the output mode. Possible values for LIBRARY are:\n"
             "\n"
             "hermes\n"
             "    Generate code that can be used with the Hermes library."
    )

    parser.add_argument(
        "--std",
        choices=["c++17"],
        dest="output_lang",
        metavar="LANG",
        default="c++17",
        help="Determine the output language. Possible values for LANG are:\n"
             "\n"
             "c++17\n"
             "    2017 ISO C++ standard."
    )

    parser.add_argument(
        "--c_out",
        type=Path,
        metavar="OUT_DIR",
        dest="impl_output_dir",
        help="The directory where C/C++ implementation files should be generated.",
        default=Path.cwd()
    )

    parser.add_argument(
        "--h_out",
        type=Path,
        metavar="OUT_DIR",
        dest="header_output_dir",
        help="The directory where C/C++ header files should be generated.",
        default=Path.cwd()
    )

    parser.add_argument(
        "--copyright-file",
        type=Path,
        default=None,
        help="Use the text in COPYRIGHT_FILE as the copyright header of all generated files."
    )

    parser.add_argument(
        "--color-diagnostics",
        choices=["never", "always", "auto"],
        metavar="WHEN",
        default="auto",
        help="Use escape sequences in diagnostic messages to display them in color in the terminal. WHEN is never, "
             "always, or auto (the default)."
    )

    parser.add_argument(
        "--debug",
        "-d",
        action="store_const",
        dest="loglevel",
        const=logging.DEBUG,
        default=logging.WARNING,
        help=argparse.SUPPRESS
    )

    parser.add_argument(
        "--verbose",
        "-v",
        action=_SetVerbosity,
        dest="loglevel",
        default=logging.INFO,
        help="Increase verbosity"
    )

    parser.add_argument(
        "--version",
        "-V",
        action=_PrintVersion,
        help="Display rpcc version information."
    )

    return parser.parse_args(args)



# Copyright 2021-2022, Barcelona Supercomputing Center (BSC), Spain
#
# This software was partially supported by the EuroHPC-funded project ADMIRE
#   (Project ID: 956748, https://www.admire-eurohpc.eu).
#
# This file is part of rpcc.
#
# rpcc is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
#
# rpcc is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License along with rpcc.  If not, see
# <https://www.gnu.org/licenses/>.
#
#  SPDX-License-Identifier: GPL-3.0-or-later

import sys
from rpcc import Rpcc
from rpcc.arguments import parse_args
from rpcc.exceptions import RpccSyntaxError
from rpcc.console import console
from loguru import logger


def main(args=None):
    """"The main routine"""

    if args is None:
        args = sys.argv[1:]

    args = parse_args(args)

    # configure console
    console.configure(args.color_diagnostics)

    # configure logging
    logger.remove()
    logger.add(sys.stderr, level=args.loglevel)

    try:
        Rpcc(
            args.rpc_proto_file,
            args.copyright_file,
            args.output_lang,
            args.header_output_dir,
            args.impl_output_dir
        ).transform()
    except (RpccSyntaxError, OSError) as e:
        console.eprint(f"ERROR: {e}")
        return 1

    return 0


if __name__ == "__main__":
    sys.exit(main())



# Copyright 2021-2022, Barcelona Supercomputing Center (BSC), Spain
#
# This software was partially supported by the EuroHPC-funded project ADMIRE
#   (Project ID: 956748, https://www.admire-eurohpc.eu).
#
# This file is part of rpcc.
#
# rpcc is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
#
# rpcc is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License along with rpcc.  If not, see
# <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from rich.highlighter import RegexHighlighter


class DiagnosticHighlighter(RegexHighlighter):
    """Apply a style to diagnostic messages"""

    base_style = "diagnostic."
    highlights = [
        r"(?P<location>.*?:\d+:\d+:)\s(?P<error>error:)\s(?P<message>.*?)(?:’(?P<ident>.*?)’)?(?P<context>(?:\n.*){2})(?P<caret>\^)\n",
        r"(?P<location>.*?:\d+:\d+:)\s(?P<note>note:)\s(?:’(?P<ident>.*?)’)?(?P<message>.*?)(?P<context>(?:\n.*){2})(?P<caret>\^)\n",
        r"Expected:(?:\n\s+’(?P<token>.*?)’)+"
    ]



# Copyright 2021-2022, Barcelona Supercomputing Center (BSC), Spain
#
# This software was partially supported by the EuroHPC-funded project ADMIRE
#   (Project ID: 956748, https://www.admire-eurohpc.eu).
#
# This file is part of rpcc.
#
# rpcc is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
#
# rpcc is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License along with rpcc.  If not, see
# <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import pathlib
from rpcc.transformers import FileTransformer, FileTransformerFactory


class Rpcc:
    """The main class for the DSL compiler."""
    transformer: FileTransformer

    def __init__(self, input_file: pathlib.Path, copyright_file: pathlib.Path, output_language: str,
                 header_outdir: pathlib.Path, impl_outdir: pathlib.Path) -> None:
        self.transformer = FileTransformerFactory.create_transformer(
            output_language,
            input_file,
            copyright_file,
            header_outdir,
            impl_outdir)

    def transform(self) -> None:
        self.transformer.transform()



# Copyright 2021-2022, Barcelona Supercomputing Center (BSC), Spain
#
# This software was partially supported by the EuroHPC-funded project ADMIRE
#   (Project ID: 956748, https://www.admire-eurohpc.eu).
#
# This file is part of rpcc.
#
# rpcc is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
#
# rpcc is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License along with rpcc.  If not, see
# <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import pathlib


class FilePosition:
    """An object representing a specific location in the parsed text"""

    def __init__(self, filename: pathlib.Path, text: str, line: int, column: int, pos_in_stream: int) -> None:
        """Initialize a `FileLocation`.

        :param filename: The filename.
        :param text: The parsed text.
        :param line: The line of interest.
        :param column: The column of interest.
        :param pos_in_stream: The position of the location of interest in the file stream.
        """

        self.filename = filename
        self.text = text
        self.line = line
        self.column = column
        self.pos_in_stream = pos_in_stream



# Copyright 2021-2022, Barcelona Supercomputing Center (BSC), Spain
#
# This software was partially supported by the EuroHPC-funded project ADMIRE
#   (Project ID: 956748, https://www.admire-eurohpc.eu).
#
# This file is part of rpcc.
#
# rpcc is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
#
# rpcc is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License along with rpcc.  If not, see
# <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from typing import Collection

from rpcc.meta import FilePosition


class ConfigurationError(ValueError):
    pass


def assert_config(value, options: Collection, msg="Got %r, expected one of %s"):
    if value not in options:
        raise ConfigurationError(msg % (value, options))


class RpccSyntaxError(SyntaxError):
    label = "syntax error"

    def __init__(self, location: FilePosition) -> None:
        self.location = location

    def __str__(self):
        return f"{self.location.filename}:{self.location.line}:{self.location.column}: error: {self.label}"


class UnexpectedToken(RpccSyntaxError):
    label = "unexpected token"

    def __init__(self, location: FilePosition, token, expected, terminals):
        super().__init__(location)
        self.token = token
        self.expected_tokens = expected
        self.terminals = terminals

    @staticmethod
    def _format_expected(expected, terminals):
        d = {t.name: t for t in terminals}

        expected = [str(d[t_name].pattern).replace('\\', "").replace('\'', '’') if t_name in d else t_name for t_name in
                    expected]

        if len(expected) > 1:
            return "Expected one of:\n\t* %s\n" % '\n\t* '.join(expected)
        else:
            return "Expected:\n\t%s\n" % '\n\t* '.join(expected)

    def __str__(self):
        return (super().__str__() + f" ’{self.token}’\n" +
                f"{get_context(self.location)}\n" +
                f"{self._format_expected(self.expected_tokens, self.terminals)}")


class UnexpectedCharacters(RpccSyntaxError):
    label = "unexpected character"

    def __str__(self):
        return (super().__str__() + "\n" +
                get_context(self.location))


class UnexpectedEOF(RpccSyntaxError):
    label = "unexpected End-of-File."


class MissingRpcDefinition(RpccSyntaxError):
    label = "missing rpc definition"

    def __str__(self):
        return (super().__str__() + "\n" +
                get_context(self.location))


class MissingRpcName(RpccSyntaxError):
    label = "missing rpc name"

    def __str__(self):
        return (super().__str__() + "\n" +
                get_context(self.location))


class EmptyRpcDefinition(RpccSyntaxError):
    label = "rpc definition is empty"

    def __str__(self):
        return (super().__str__() + "\n" +
                get_context(self.location))


class RpcRedefinition(RpccSyntaxError):
    label = "redefinition of rpc"
    prev_label = "previously defined here"

    def __init__(self, location: FilePosition, name: str, prev_definition: FilePosition):
        super().__init__(location)
        self.name = name
        self.prev = prev_definition

    def __str__(self):
        return (super().__str__() + f" ’{self.name}’\n" +
                get_context(self.location) + "\n" +
                f"{self.prev.filename}:{self.prev.line}:{self.prev.column}: note: ’{self.name}’ {self.prev_label}\n" +
                get_context(self.prev))


def get_context(location: FilePosition, span: int = 40) -> str:
    """Returns a pretty string pinpointing the error in the text,
       with span amount of context characters around it.
    """
    pos = location.pos_in_stream
    start = max(pos - span, 0)
    end = pos + span
    max_width = 6
    if not isinstance(location.text, bytes):
        before = location.text[start:pos].rsplit('\n', 1)[-1]
        after = location.text[pos:end].split('\n', 1)[0]
        return f'{location.line:>{max_width}} | ' + before + after + '\n' + \
               ' ' * max_width + ' | ' + ' ' * len(before.expandtabs()) + '^\n'
    else:
        before = location.text[start:pos].rsplit(b'\n', 1)[-1]
        after = location.text[pos:end].split(b'\n', 1)[0]
        return (before + after + b'\n' + b' ' * len(before.expandtabs()) + b'^\n').decode("ascii",
                                                                                          "backslashreplace")



# Copyright 2021-2022, Barcelona Supercomputing Center (BSC), Spain
#
# This software was partially supported by the EuroHPC-funded project ADMIRE
#   (Project ID: 956748, https://www.admire-eurohpc.eu).
#
# This file is part of rpcc.
#
# rpcc is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
#
# rpcc is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License along with rpcc.  If not, see
# <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import itertools
from abc import ABC, abstractmethod
from typing import List, Any, Optional


class Argument:
    """A remote procedure argument.

    Parameters:
        id: a string with the argument's name.
        typeinfo: an object instance with the argument's type information.

    Example:
        >>> Argument("foobar", Integer(32, True))
        Argument(...)
    """

    id: str
    typeinfo: Any

    def __init__(self, id: str, typeinfo: Any) -> None:
        self.id = id
        self.typeinfo = typeinfo

    def __repr__(self) -> str:
        return f"Argument(id='{self.id}', typeinfo='{self.typeinfo}')"


class ReturnValue:
    """A remote procedure return value.

    Parameters:
        id: a string with the return variable's name.
        typeinfo: an object instance with the argument's type information.

    Example:
        >>> ReturnValue("barbaz", Integer(32, True))
        ReturnValue(...)
    """

    id: str
    typeinfo: Any

    def __init__(self, id: str, typeinfo: Any) -> None:
        self.id = id
        self.typeinfo = typeinfo

    def __repr__(self) -> str:
        return f"ReturnVariable(id='{self.id}', typeinfo='{self.typeinfo}')"


class ArgumentList:
    """A list of RPC arguments."""

    args: List[Argument]

    def __init__(self, args: List[Argument]):
        self.args = args

    def __iter__(self):
        for arg in self.args:
            yield arg

    def __len__(self):
        return len(self.args)


class ReturnValueList:
    """A list of RPC return values."""

    retvals: List[ReturnValue]

    def __init__(self, retvals: List[ReturnValue]):
        self.retvals = retvals

    def __iter__(self):
        for retval in self.retvals:
            yield retval

    def __len__(self):
        return len(self.retvals)


class ConditionalDefinition(ABC):
    @property
    @abstractmethod
    def start_keyword(self):
        raise NotImplemented

    @property
    @abstractmethod
    def end_keyword(self):
        raise NotImplemented

    @property
    @abstractmethod
    def symbol(self):
        raise NotImplemented

    @symbol.setter
    @abstractmethod
    def symbol(self, value):
        raise NotImplemented


class RemoteProcedure:
    """A remote procedure.

    Parameters:
        name: a string with the remote procedure's name.
        args: an ArgumentList containing the remote procedure's arguments.
        retvals: a ReturnValueList containing the remote procedure's return variables.

    Example:
        >>> RemoteProcedure("send_message",
        >>>                 ArgumentList([Argument("message", "string")],
        >>>                 ReturnValueList([ReturnValue("retval", "uint32")])))
        RemoteProcedure(...)
    """

    id: int
    name: str
    args: ArgumentList
    retvals: ReturnValueList
    include_if_expr: Optional[ConditionalDefinition]

    id_iter = itertools.count()

    def __init__(self, id: int, name: str, args: ArgumentList, retvals: ReturnValueList,
                 include_if_expr: Optional[ConditionalDefinition]) -> None:
        self.id = id
        self.name = name
        self.args = args
        self.retvals = retvals
        self.include_if_expr = include_if_expr

    def __repr__(self) -> str:
        return f"RemoteProcedure(id={self.id}, name='{self.name}', args={self.args}, retvals={self.retvals})"

    @staticmethod
    def generate_id(margo_compatible: bool = False) -> int:
        if not margo_compatible:
            return next(RemoteProcedure.id_iter)
        raise NotImplemented
        # TODO: generate a margo compatible id using the rpc name


class IRTree:
    """The Intermediate Representation Tree for our source-to-source compiler.

    Parameters:
        rpcs: a list of the remote procedures defined in the input file.

    Example:
        >>> IRTree([
        >>>    RemoteProcedure("send_message",
        >>>                    ArgumentList([Argument("message", "string")]),
        >>>                    ReturnValueList([ReturnValue("retval", "uint32"))])
        >>>    RemoteProcedure("shutdown", ArgumentList([]), ReturnValueList([ReturnValue("retval", "uint32")]))
        >>> ]
        Tree(...)
    """

    package: Optional[str]
    rpcs: List[RemoteProcedure]

    def __init__(self, package: Optional[str], rpcs: List[RemoteProcedure]):
        self.package = package
        self.rpcs = rpcs



# Copyright 2021-2022, Barcelona Supercomputing Center (BSC), Spain
#
# This software was partially supported by the EuroHPC-funded project ADMIRE
#   (Project ID: 956748, https://www.admire-eurohpc.eu).
#
# This file is part of rpcc.
#
# rpcc is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
#
# rpcc is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License along with rpcc.  If not, see
# <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from typing import Any

from rich.console import Console as RichConsole

__all__ = ["console"]

from rpcc.console.highlighter import DiagnosticHighlighter
from rpcc.console.theme import diagnostics_theme


class Console:
    """The main Console class."""

    def __init__(self) -> None:
        """Initialize a `Console` with default options."""
        self.console = RichConsole(highlight=False)
        self.error_console = RichConsole(stderr=True)

    def configure(self, colorize_output: str = "auto") -> None:
        """Reconfigure a `Console`.

        :param colorize_output: Control whether the `Console` should emit ANSI color codes when printing messages.
        The argument accepts either ``"always"``, ``"never"`` or ``"auto"`` (the default).
        """
        ft = {
            "always": True,
            "never": False,
            "auto": None
        }.get(colorize_output, None)

        self.console = RichConsole(force_terminal=ft, highlight=False)
        self.error_console = RichConsole(stderr=True,
                                         theme=diagnostics_theme,
                                         highlighter=DiagnosticHighlighter(),
                                         force_terminal=ft)

    def print(self, *args: Any):
        """Print a message to `stdout`.

        :param args: Objects to print to the terminal.
        """
        self.console.print(*args)

    def eprint(self, *args: Any):
        """Print a message to `stderr`.

        :param args: Objects to print to the terminal.
        """
        self.error_console.print(*args)


# the default console
console = Console()


# ---------------------------------------------------------------------------
# Utility functions at module level.
# Basically delegate everything to `console`
# ---------------------------------------------------------------------------
def configure(colorize: str = "auto") -> None:
    """Configure the default `console`.

    :param colorize: Control whether the default console should emit ANSI color codes when printing messages.
    The argument accepts either ``"always"``, ``"never"`` or ``"auto"`` (the default).
    """
    console.configure(colorize)


def print(*args: Any) -> None:
    """Print to `stdout` using the default console.

    :param args: The objects to print.
    """
    console.print(*args)


def eprint(*args: Any):
    """Print to `stderr` using the default console.

    :param args: The objects to print.
    """
    console.eprint(*args)



# Copyright 2021-2022, Barcelona Supercomputing Center (BSC), Spain
#
# This software was partially supported by the EuroHPC-funded project ADMIRE
#   (Project ID: 956748, https://www.admire-eurohpc.eu).
#
# This file is part of rpcc.
#
# rpcc is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
#
# rpcc is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License along with rpcc.  If not, see
# <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import pathlib
from typing import Tuple

from lark import Lark, Tree
import lark.exceptions
from loguru import logger

from rpcc.exceptions import (
    EmptyRpcDefinition, MissingRpcDefinition, MissingRpcName, UnexpectedToken,
    UnexpectedCharacters, UnexpectedEOF
)
from rpcc.meta import FilePosition

INVALID_INPUT_EXAMPLES = {
    EmptyRpcDefinition: ['rpc foo {}',
                         'rpc foo { arguments {} returns {} } rpc bar {}'],
    MissingRpcDefinition: ['',
                           'foo',
                           'foo {',
                           'foo {}',
                           '{',
                           'arguments {',
                           'returns {'
                           ],
    MissingRpcName: ['rpc {']
}


class Parser:
    """The main parser class"""

    def __init__(self) -> None:
        self.parser = Lark.open_from_package("rpcc", "grammars/rpcc.lark", propagate_positions=True, parser='lalr')

    def parse(self, input_file: pathlib.Path) -> Tuple[Tree, str]:
        """Parse an input file.

        :param input_file: A `pathlib.Path` containing the path to the RPC description file to parse.
        :return: An AST tree with a representation of the parsed text.
        """

        try:
            file = open(input_file, "r", encoding="utf-8")
        except (FileNotFoundError, EnvironmentError) as err:
            logger.error(f"Error parsing file: {err}")
            raise
        else:
            with file:
                text = file.read()
                try:
                    return self.parser.parse(text), text
                except lark.exceptions.UnexpectedInput as u:

                    location = FilePosition(input_file, text, u.line, u.column, u.pos_in_stream)

                    exc_class = u.match_examples(self.parser.parse, INVALID_INPUT_EXAMPLES, use_accepts=True)

                    if exc_class:
                        raise exc_class(location) from None

                    if isinstance(u, lark.exceptions.UnexpectedToken):
                        raise UnexpectedToken(location, u.token, u.expected, self.parser.terminals) from None

                    if isinstance(u, lark.exceptions.UnexpectedCharacters):
                        raise UnexpectedCharacters(location) from None

                    if isinstance(u, lark.exceptions.UnexpectedEOF):
                        raise UnexpectedEOF(location) from None

                    raise u



# Copyright 2021-2022, Barcelona Supercomputing Center (BSC), Spain
#
# This software was partially supported by the EuroHPC-funded project ADMIRE
#   (Project ID: 956748, https://www.admire-eurohpc.eu).
#
# This file is part of rpcc.
#
# rpcc is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
#
# rpcc is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License along with rpcc.  If not, see
# <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from abc import ABCMeta, abstractmethod
from typing import Optional, Dict, Type, Callable


class NameSpace:
    id: str

    def __init__(self, id='') -> None:
        self.id = id

    def __str__(self) -> str:
        return self.id

    def __repr__(self) -> str:
        return f"NameSpace(id='{self.id}')"


class CxxType(metaclass=ABCMeta):
    @abstractmethod
    def is_fundamental(self) -> bool:
        raise NotImplemented

    @abstractmethod
    def member_operator(self) -> str:
        raise NotImplemented

    @abstractmethod
    def hg_type(self) -> str:
        raise NotImplemented


class FundamentalType(CxxType, metaclass=ABCMeta):
    def is_fundamental(self) -> bool:
        return True

    def member_operator(self) -> str:
        raise ValueError("C++ fundamental types don't have member operators!")

    def hg_type(self) -> str:
        raise NotImplemented

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}()"


class CompoundType(CxxType, metaclass=ABCMeta):
    classname: str
    namespace: Optional[NameSpace]

    def __init__(self, classname: str, namespace: Optional[NameSpace] = NameSpace()) -> None:
        self.classname = classname
        self.namespace = namespace

    def is_fundamental(self) -> bool:
        return False

    def member_operator(self) -> str:
        return "."

    def hg_type(self) -> str:
        raise NotImplemented

    def __str__(self) -> str:
        return '::'.join(filter(None, [str(self.namespace), self.classname]))

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(classname='{self.classname}', namespace={self.namespace.__repr__()})"


class Void(FundamentalType):
    def __str__(self) -> str:
        return "void"

    def hg_type(self) -> str:
        return "void"


class Bool(FundamentalType):
    def __str__(self) -> str:
        return "bool"

    def hg_type(self) -> str:
        return "hg_bool_t"


class Integer(FundamentalType):
    width: int
    unsigned: bool

    def __init__(self, width: int, unsigned: bool) -> None:
        self.width = width
        self.unsigned = unsigned

    def hg_type(self) -> str:
        return f"hg_{self}"

    def __str__(self) -> str:
        prefix = "u" if self.unsigned else ""
        return f"{prefix}int{self.width}_t"

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(width={self.width}, unsigned={self.unsigned})"


class CSize(FundamentalType):
    def hg_type(self) -> str:
        return "hg_size_t"

    def __str__(self) -> str:
        return "size_t"


class Float(FundamentalType):
    def hg_type(self) -> str:
        return str(self)

    def __str__(self) -> str:
        return "float"


class Double(FundamentalType):
    def hg_type(self) -> str:
        return str(self)

    def __str__(self) -> str:
        return "double"


class Object(CompoundType):
    """A generic object."""
    pass


class String(CompoundType):
    def __init__(self) -> None:
        super().__init__("string", NameSpace("std"))

    def hg_type(self) -> str:
        return "hg_const_string_t"


class ExposedBuffer(CompoundType):
    def __init__(self) -> None:
        super().__init__("exposed_memory", NameSpace("hermes"))

    def hg_type(self) -> str:
        return "hg_bulk_t"


class Pointer(CompoundType):
    to_type: CxxType

    def __init__(self, to_type: CxxType) -> None:
        self.to_type = to_type

    def member_operator(self) -> str:
        return "->"

    def __str__(self) -> str:
        return f"{self.to_type} *"


class ReferenceType(CompoundType):
    to_type: CxxType

    def __init__(self, to_type: CxxType) -> None:
        self.to_type = to_type

    def __str__(self) -> str:
        return f"{self.to_type}&"


class Identifier:
    id: str
    typeinfo: CxxType
    parent: Optional["Identifier"]

    def __init__(self, id: str, typeinfo: CxxType, parent: Optional["Identifier"] = None) -> None:
        self.id = id
        self.typeinfo = typeinfo
        self.parent = parent

    def member_operator(self) -> str:
        return self.typeinfo.member_operator()

    def as_declarator(self) -> str:
        return f"{self.typeinfo} {self.id}"

    def as_member_of(self) -> str:
        if not self.parent:
            return str(self.id)
        return f"{self.parent}{self.parent.member_operator()}{self.id}"

    def as_hg_conversion(self) -> str:

        valid_hg_conversions: Dict[Type[CxxType], Callable[[Identifier], str]] = {
            String: lambda a: f"{a}.c_str()",
            ExposedBuffer: lambda a: f"hg_bulk_t({a})",
        }

        if self.typeinfo.is_fundamental():
            return self.id

        if type(self.typeinfo) in valid_hg_conversions:
            return valid_hg_conversions[type(self.typeinfo)](self)

        raise ValueError(f"No known conversion for compound type '{self.typeinfo}'")

    def __str__(self) -> str:
        return self.id

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(" \
               f"id='{self.id}', typeinfo={self.typeinfo.__repr__()}, parent={self.parent.__repr__() or 'None'})"


class LValueReference(Identifier):

    def __init__(self, id: str, typeinfo: CxxType, parent: Optional["Identifier"] = None) -> None:
        super().__init__(id, typeinfo, parent)

    def as_declarator(self) -> str:
        return f"{self.typeinfo}& {self.id}"


class ConstLValueReference(Identifier):

    def __init__(self, id: str, typeinfo: CxxType, parent: Optional["Identifier"] = None) -> None:
        super().__init__(id, typeinfo, parent)

    def as_declarator(self) -> str:
        return f"const {self.typeinfo}& {self.id}"


class RValueReference(Identifier):

    def __init__(self, id: str, typeinfo: CxxType, parent: Optional["Identifier"] = None) -> None:
        super().__init__(id, typeinfo, parent)

    def as_declarator(self) -> str:
        return f"{self.typeinfo}&& {self.id}"



import pathlib
from setuptools import setup, find_packages
from rpcc.version import __version__

# package metadata
NAME = "rpcc"
DESCRIPTION = "A compiler for Mercury-based RPCs"
URL = "https://storage.bsc.es/projects/rpcc"
AUTHOR = "Alberto Miranda"
EMAIL = "alberto.miranda@bsc.es"

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

setup(
    name=NAME,
    version=__version__,
    description=DESCRIPTION,
    long_description=README,
    long_description_content_type="text/markdown",
    url=URL,
    author=AUTHOR,
    author_email=EMAIL,
    license="GPLv3",
    python_requires='>=3.6.0',
    packages=find_packages(exclude=("tests",)),
    include_package_data=True,
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Topic :: Software Development :: Code Generators"
    ],
    entry_points={
        "console_scripts": [
            "rpcc = rpcc.__main__:main"
        ]
    },
)



# Copyright 2021-2022, Barcelona Supercomputing Center (BSC), Spain
#
# This software was partially supported by the EuroHPC-funded project ADMIRE
#   (Project ID: 956748, https://www.admire-eurohpc.eu).
#
# This file is part of rpcc.
#
# rpcc is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
#
# rpcc is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License along with rpcc.  If not, see
# <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import pathlib
from typing import List, Tuple, Optional, Dict

import lark
from lark import v_args, Discard
from loguru import logger

import rpcc.parser
from rpcc.exceptions import RpcRedefinition
from rpcc.transformers.cxx17.formatter import FileFormatter
from rpcc.meta import FilePosition
from rpcc.parser import Parser
from rpcc.ir import RemoteProcedure, Argument, ReturnValue, IRTree, ArgumentList, ReturnValueList, ConditionalDefinition
from rpcc.transformers import FileTransformer
from .types import Bool, Double, Float, Integer, String, ExposedBuffer, CSize


class _RpcInfo:
    pos_in_text: FilePosition

    def __init__(self, pos_in_text: FilePosition):
        self.pos_in_text = pos_in_text


class _CxxCompilerDefines(ConditionalDefinition):
    _symbol: str

    def __init__(self, symbol):
        self._symbol = symbol

    @property
    def start_keyword(self):
        return "#ifdef"

    @property
    def end_keyword(self):
        return "#endif"

    @property
    def symbol(self):
        return self._symbol

    @symbol.setter
    def symbol(self, value):
        self._symbol = value


class _CxxCompilerNotDefines(ConditionalDefinition):
    _symbol: str

    def __init__(self, symbol):
        self._symbol = symbol

    @property
    def start_keyword(self):
        return "#ifndef"

    @property
    def end_keyword(self):
        return "#endif"

    @property
    def symbol(self):
        return self._symbol

    @symbol.setter
    def symbol(self, value):
        self._symbol = value


class _AstTransformer(lark.Transformer):
    input_file: pathlib.Path
    text: str
    rpc_info: Dict[str, _RpcInfo]

    def __init__(self, input_file: pathlib.Path, text: str):
        super().__init__()
        self.input_file = input_file
        self.text = text
        self.rpc_info = dict()

    @v_args(inline=True)
    def start(self, package: Optional[str], rpcs: List[RemoteProcedure]) -> IRTree:
        """Transform the AST tree generated by Lark into an `IR` instance, that we can use to generate the output files
        """

        return IRTree(package, rpcs)

    @v_args(inline=True)
    def package(self, name: str) -> str:
        return name

    # rpc_list can be converted directly into list
    rpc_list = list

    @v_args(inline=True, meta=True)
    def rpc(self, meta: lark.tree.Meta, name: str, id: Optional[int], include_if_expr: Optional[ConditionalDefinition],
            args: Optional[ArgumentList], retvals: Optional[ReturnValueList]) -> RemoteProcedure:
        """Transform a `rpc` node from the AST into a `RemoteProcedure` object from a name, a list of `Argument`s and a
        `ReturnVariable`.

        :param meta: Metainformation about the current Token such as line, column, start_pos, etc.
        :param name: A name for the remote procedure.
        :param id: An optional numeric id the remote procedure.
        :param args: An `ArgumentList` describing the input arguments for the remote procedure. `ArgumentList` may be
        empty if the input file defined an empty `arguments` clause, or `None` if not defined.
        :param retvals: A `ReturnValueList` describing the remote procedure's return values. `ReturnValueList` may be
        empty if the input file defined an empty `returns` clause, or `None` if not defined.
        :param include_if_expr: A C/C++ preprocessor expression specifying when this rpc should be included in
        compilation.
        :return: A `RemoteProcedure` object describing the remote procedure.
        """

        if not id:
            id = RemoteProcedure.generate_id(margo_compatible=False)

        if not args:
            args = ArgumentList([])

        if not retvals:
            retvals = ReturnValueList([])

        pos = FilePosition(self.input_file, self.text, meta.line, meta.column, meta.start_pos)
        if name in self.rpc_info:
            prev_pos = self.rpc_info[name].pos_in_text
            raise RpcRedefinition(pos, name, prev_pos)
        else:
            self.rpc_info[name] = _RpcInfo(pos)

        return RemoteProcedure(id, name, args, retvals, include_if_expr)

    @v_args(inline=True)
    def rpc_name(self, name: str) -> str:
        """Transform a `rpc_name` node from the AST into its corresponding string. The string is extracted from the AST
        nodes' value.

        :param name: A string value from the `rpc_name` node.
        :return: A name for the remote procedure.
        """
        return name

    @v_args(inline=True)
    def rpc_id(self, id: int) -> int:
        """Transform a `rpc_id` node from the AST into its corresponding id. The id is extracted from the AST
        nodes' value.

        :param id: An integer value from the `rpc_id` node.
        :return: An id for the remote procedure.
        """
        return id

    @v_args(inline=True)
    def include_if(self, expr) -> ConditionalDefinition:
        return expr

    @v_args(inline=True)
    def ifdef(self, symbol) -> ConditionalDefinition:
        return _CxxCompilerDefines(symbol)

    @v_args(inline=True)
    def ifndef(self, symbol) -> ConditionalDefinition:
        return _CxxCompilerNotDefines(symbol)

    def rpc_args(self, children: List[Tuple[str, str]]) -> ArgumentList:
        """Transform a list of tuples returned by Lark into a list of `rpc.Arguments`. The tuples are implicitly
        constructed by Lark by transforming all the `var` nodes in the AST that have this particular `rpc_args` node
        as a parent.

        :param children: A list of tuples describing the rpc arguments.
        :return: A list of `rpc.Arguments`
        """
        return ArgumentList([Argument(id, typeinfo) for (typeinfo, id) in children])

    def rpc_retvals(self, children: List[Tuple[str, str]]) -> ReturnValueList:
        """Transform a tuple returned by Lark into a `ReturnVariable`. The tuple is implictly constructed by Lark by
        transforming the `var` node under this particular `rpc_return` node.

        :param children: A list of tuples describing the rpc return values.
        :return: A list of `rpc.ReturnVariables`
        """
        return ReturnValueList([ReturnValue(id, typeinfo) for (typeinfo, id) in children])

    # variables can be converted directly into tuples
    var = tuple

    def bool(self, _) -> Bool:
        return Bool()

    def double(self, _) -> Double:
        return Double()

    def float(self, _) -> Float:
        return Float()

    def int8(self, _) -> Integer:
        return Integer(8, False)

    def int16(self, _) -> Integer:
        return Integer(16, False)

    def int32(self, _) -> Integer:
        return Integer(32, False)

    def int64(self, _) -> Integer:
        return Integer(64, False)

    def uint8(self, _) -> Integer:
        return Integer(8, True)

    def uint16(self, _) -> Integer:
        return Integer(16, True)

    def uint32(self, _) -> Integer:
        return Integer(32, True)

    def uint64(self, _) -> Integer:
        return Integer(64, True)

    def csize(self, _) -> CSize:
        return CSize()

    def string(self, _) -> String:
        return String()

    def exposed_buffer(self, _) -> ExposedBuffer:
        return ExposedBuffer()

    # variable names can be coverted directly into strings
    NAME = str
    DOTTED_NAME = str

    def EOS(self, children):
        return Discard


class Cxx17FileTransformer(FileTransformer):
    input_file: pathlib.Path
    copyright_file: pathlib.Path
    header_outdir: pathlib.Path
    impl_outdir: pathlib.Path
    parser: rpcc.parser.Parser
    ast_transformer: _AstTransformer

    def _get_output_paths(self) -> Tuple[pathlib.Path, pathlib.Path]:
        prefix = self.input_file.stem.replace('.', '_')
        return (pathlib.Path(self.header_outdir, f"{prefix}-rpcc.hpp"),
                pathlib.Path(self.impl_outdir, f"{prefix}-rpcc.cpp"))

    def __init__(self, input_file: pathlib.Path, copyright_file: pathlib.Path, header_outdir: pathlib.Path,
                 impl_outdir: pathlib.Path):
        self.input_file = input_file
        self.copyright_file = copyright_file
        self.header_outdir = header_outdir
        self.impl_outdir = impl_outdir
        self.parser = Parser()

    def transform(self) -> None:

        ast, text = self.parser.parse(self.input_file)

        try:
            ir = _AstTransformer(self.input_file, text).transform(ast)
        except lark.exceptions.VisitError as e:
            raise e.orig_exc

        header_file, impl_file = self._get_output_paths()
        header_text, impl_text = FileFormatter(self.input_file, self.copyright_file, header_file, impl_file).format(ir)

        try:
            with open(header_file, "w") as f:
                f.write(header_text)
        except OSError as e:
            raise OSError(f"Error writing header file:\n  {e}") from None

        try:
            with open(impl_file, "w") as f:
                f.write(impl_text)
        except OSError as e:
            raise OSError(f"Error writing implementation file:\n  {e}") from None
